"""Gateway connecting to ATAG thermostat."""
import aiohttp
import logging
from datetime import datetime, timedelta
import re
import uuid
import asyncio
import logging
from .errors import raise_error
from .entities import Report, Climate, DHW

_LOGGER = logging.getLogger(__name__)


class AtagOne:
    """Central data store entity."""

    def __init__(self, host, session, device=None, email=None, port=10000):
        """Initialize main AtagOne object."""
        self.host = host
        self.email = email or ""
        self.port = port
        if email is None:
            _LOGGER.debug("No email address provided.")
        self.session = session
        self._device = device
        self._authorized = device is not None
        self._mac = ":".join(re.findall("..", "%012x" % uuid.getnode()))
        self._last_call = datetime(1970, 1, 1)
        self._lock = asyncio.Lock()

        self.climate = None
        self.dhw = None
        self.report = None

    @property
    def id(self):
        """Return the ID of the bridge."""
        if self.report:
            return self.report["device_id"].state
        return self._device

    @property
    def apiversion(self):
        """Return the ID of the bridge."""
        if self.report:
            return self.report["download_url"].state

    async def authorize(self):
        """Check auth status."""
        json = {
            "pair_message": {
                "seqnr": 1,
                "account_auth": {"user_account": self.email, "mac_address": self._mac},
                "accounts": {
                    "entries": [
                        {
                            "user_account": self.email,
                            "mac_address": self._mac,
                            "device_name": "HomeAssistant",
                            "account_type": 1,
                        }
                    ]
                },
            }
        }
        await self.request("post", "pair", json)
        self.authorized = True
        _LOGGER.debug(f"Authorized: {self.authorized}")
        return self.authorized

    async def request(self, meth, path, json=None):
        """Make a request to the API."""
        url = f"http://{self.host}:{self.port}/{path}"
        async with self._lock:
            slept = (datetime.utcnow() - self._last_call).total_seconds()
            if slept < 5:
                await asyncio.sleep(5 - slept)
            self._last_call = datetime.utcnow()
            try:
                async with self.session.request(meth, url, json=json) as res:
                    data = await res.json()
                    _raise_on_error(data)
                    return data
            except aiohttp.ClientConnectorError as err:
                raise_error(err, 2)
            except Exception as err:
                raise_error(err, 5)

    async def update(self, info=71, force=False):
        """Get latest data from API."""
        if (datetime.utcnow() - self._last_call).total_seconds() > 15 or force:
            json = {
                "retrieve_message": {
                    "seqnr": 1,
                    "account_auth": {
                        "user_account": self.email,
                        "mac_address": self._mac,
                    },
                    "info": info,
                }
            }
            res = await self.request("get", "retrieve", json)
            res = res["retrieve_reply"]
            res["report"].update(res["report"].pop("details"))
            if self.report is None:
                self.report = Report(res, self.update, self.setter)
                self.climate = Climate(self.report)
                self.dhw = DHW(self.report)
            self.report.update(res)
            return self.report
        return None

    async def setter(self, **kwargs):
        """Set control items."""
        json = {
            "update_message": {
                "seqnr": 1,
                "account_auth": {"user_account": self.email, "mac_address": self._mac},
                "control": {},
                "configuration": {},
            }
        }

        for key, val in kwargs.items():
            json["update_message"]["control"][key] = val
            if key == "ch_mode" and val == 3:
                json["update_message"]["control"]["vacation_duration"] = int(
                    timedelta(days=1).total_seconds()
                )
                json["update_message"]["configuration"]["start_vacation"] = int(
                    (datetime.utcnow() - datetime(2000, 1, 1)).total_seconds()
                )
        res = await self.request("post", "update", json)
        return res["update_reply"]


def _raise_on_error(data):
    """Check response for error message."""
    if data[list(data.keys())[0]]["acc_status"] != 2:
        raise_error(data, 1)

    if isinstance(data, list):
        data = data[0]

    if isinstance(data, dict) and "error" in data:
        raise_error(data)



"""Error handling for atag_Thermostat."""


class AtagException(Exception):
    """Base error for AtagOne devices."""


class Unauthorized(AtagException):
    """Failed to authenticate."""


class RequestError(AtagException):
    """Unable to fulfill request."""


class ResponseError(AtagException):
    """Invalid response."""


class Response404Error(AtagException):
    """Invalid response."""


class UnknownAtagError(AtagException):
    """Invalid response."""


ERRORS = {
    1: Unauthorized,
    2: RequestError,
    3: ResponseError,
    4: Response404Error,
    5: UnknownAtagError,
}


def raise_error(error, type=None):
    """Raise the appropriate error."""
    cls = ERRORS.get(type, AtagException)
    raise cls(error)



"""Classes within AtagOne object."""
from .const import CLASSES, SENSORS, STATES
from datetime import datetime, timedelta


def convert_time(seconds, sensorclass="time"):
    """Convert reported time to real time."""
    if sensorclass == "duration":
        return str(timedelta(seconds=seconds))
    return str(datetime(2000, 1, 1) + timedelta(seconds=seconds))


class Report:
    """Main object to hold report and control data."""

    def __init__(self, data, update, setter):
        """Initiate object for Sensor and Control data."""
        self._update = update
        self._setter = setter
        self._items = {}
        self._data = data
        self._classes = CLASSES
        self._classes["temp"][1] = STATES["temp_unit"][
            self._data["configuration"]["temp_unit"]
        ]
        self._process_raw(self._data)

    def update(self, data):
        """Process latest data."""
        self._data = data
        self._process_raw(self._data)

    def _process_raw(self, raw):
        """Push data to the sensor and control objects."""
        for grp in ["configuration", "status", "report", "control"]:
            for id, raw_i in raw[grp].items():
                name = SENSORS.get(id)
                obj = self._items.get(name) or self._items.get(id)

                if obj is not None:
                    obj.raw = raw_i
                elif grp == "control":
                    self._items[name or id] = Control(
                        id, raw_i, self._classes, self._setter
                    )
                else:
                    self._items[name or id] = Sensor(id, raw_i, self._classes)

    def items(self):
        """Return the report objects."""
        return self._items.values()

    @property
    def report_time(self):
        """Return latest report time."""
        return self._items["report_time"].state

    def __getitem__(self, obj_id):
        """Return selected sensor object by name or ID."""
        if obj_id in SENSORS:
            return self._items[SENSORS[obj_id]]
        return self._items[obj_id]

    def __iter__(self):
        """Iterate over sensor and control objects."""
        return iter(self._items.values())


class Sensor:
    """Represents an Atag sensor."""

    def __init__(self, id, raw, classes):
        """Initiate sensor object."""
        self.id = id
        self.raw = raw
        self._info = classes.get(self.id.split("_")[-1])
        if id == "tout_avg":
            self._info = classes["temp"]
        self._states = STATES.get(self.id)

    @property
    def name(self):
        """Return the readable name of the Sensor."""
        return SENSORS.get(self.id) or self.id

    @property
    def sensorclass(self):
        """Return the sensorclass if known."""
        if self._info:
            return self._info[0]

    @property
    def state(self):
        """Return state, if known in readable format."""
        if self.sensorclass in ["time", "duration"]:
            return convert_time(self.raw, self.sensorclass)
        if self.id == "boiler_status":
            return {
                "burner": self.raw & 8 == 8,
                "dhw": self.raw & 4 == 4,
                "ch": self.raw & 2 == 2,
            }
        if self.id == "download_url":
            return self.raw.split("/")[-1]
        if self._states is not None:
            if isinstance(self._states[self.raw], dict):
                return self._states[self.raw]["state"]
            return self._states[self.raw]
        return self.raw

    @property
    def icon(self):
        """Return the icon corresponding to the state."""
        if self._info:
            return self._info[2]
        if self._states is not None:
            if isinstance(self._states[self.raw], dict):
                return self._states[self.raw]["icon"]
        return None

    @property
    def measure(self):
        """Return the unit of measurement if known."""
        if self._info:
            return self._info[1]
        return None

    def __repr__(self):
        """Return the name of the Sensor."""
        return self.name


class Control(Sensor):
    """Represents an Atag control."""

    def __init__(self, id, raw, classes, setter):
        """Initiate Control object."""
        super().__init__(id, raw, classes)
        self._setter = setter
        self._target = None
        self._last_call = None

    @property
    def state(self):
        """Return reported state, or target state if recently set."""
        if self._target:
            if (datetime.utcnow() - self._last_call).total_seconds() < 15:
                return self._target
            self._target = None
        if self.sensorclass in ["time", "duration"]:
            return convert_time(self.raw, self.sensorclass)
        if self._states:
            if isinstance(self._states[self.raw], dict):
                return self._states[self.raw]["state"]
            return self._states[self.raw]
        if self.id == "dhw_mode_temp":
            return self.raw % 150
        return self.raw

    async def set_state(self, target):
        """Set the Control to a new target state."""
        target_int = None
        if self._states is not None:
            target = {v.lower(): v for v in self._states.values()}.get(target.lower())
            target_int = {v: k for k, v in self._states.items()}.get(target)
        if target == self.state:
            return True
        self._target = target
        self._last_call = datetime.utcnow()
        return await self._setter(**{self.id: target_int or target})


class Climate:
    """Main climate entity."""

    def __init__(self, report):
        """Initiate the main climate object."""
        self._report = report

    @property
    def temp_unit(self):
        """Return temperature unit."""
        return self._report["temp_unit"].state

    @property
    def burnerstatus(self):
        """Return boolean for burner active."""
        return self._report["boiler_status"].state["burner"]

    @property
    def status(self):
        """Return HVAC action."""
        return self._report["boiler_status"].state["ch"]

    @property
    def flame(self):
        """Return flame level."""
        if self.status:
            return self._report["flame"].state
        return 0

    @property
    def hvac_mode(self):
        """Return the operating mode (Weather or Regular/Heat)."""
        return self._report["ch_control_mode"].state

    async def set_hvac_mode(self, target: str) -> bool:
        """Set the operating mode (Weather or Regular/Heat)."""
        await self._report["ch_control_mode"].set_state(target)

    @property
    def preset_mode(self):
        """Return the preset mode (Manual/Auto/Extend/Vacation/Fireplace)."""
        return self._report["ch_mode"].state

    @property
    def preset_mode_duration(self):
        """Return remaining time on preset mode."""
        return self._report["ch_mode_duration"].state

    async def set_preset_mode(self, target: str, **kwargs) -> bool:
        """Set the hold mode (Manual/Auto/Extend/Vacation/Fireplace)."""
        await self._report["ch_mode"].set_state(target)

    @property
    def temperature(self):
        """Return current indoor temperature."""
        return self._report["room_temp"].state

    @property
    def target_temperature(self):
        """Return target temperature."""
        return self._report["ch_mode_temp"].state

    async def set_temp(self, target: float):
        """Set target temperature."""
        await self._report["ch_mode_temp"].set_state(target)


class DHW:
    """Main Domestic Hot Water object."""

    def __init__(self, report):
        """Initiate main DHW object."""
        self._report = report

    @property
    def temp_unit(self):
        """Return temperature unit."""
        return self._report["temp_unit"].state

    @property
    def burnerstatus(self):
        """Return boolean for burner status."""
        return self._report["boiler_status"].state["burner"]

    @property
    def status(self):
        """Return boolean indicator for heating for DHW."""
        return self._report["boiler_status"].state["dhw"]

    @property
    def flame(self):
        """Return flame level if active for DHW."""
        if self.status:
            return self._report["flame"].state
        return 0

    @property
    def temperature(self):
        """Return water temperature."""
        return self._report["dhw_water_temp"].state

    @property
    def min_temp(self):
        """Return dhw min temperature."""
        return self._report["dhw_min_set"].state

    @property
    def max_temp(self):
        """Return dhw max temperature."""
        return self._report["dhw_max_set"].state

    @property
    def target_temperature(self):
        """Return dhw target temperature."""
        if self.status:
            return self._report["dhw_temp_setp"].state
        return self._report["dhw_mode_temp"].state

    async def set_temp(self, target: float):
        """Set dhw target temperature."""
        await self._report["dhw_target_temp"].set_state(target)



"""Example program to test pyatag."""
import asyncio

from pyatag import AtagOne
import aiohttp
import logging

handle = "atag"
logging.basicConfig()
_LOGGER = logging.getLogger(handle)
_LOGGER.setLevel(logging.DEBUG)


async def main():
    """Initialize session for main program."""
    async with aiohttp.ClientSession() as session:
        await run(session)


async def run(session):
    """Run example main program."""
    atag = AtagOne("localhost", session, email=None)
    await atag.authorize()
    await atag.update(force=True)

    for s in atag.report:
        _LOGGER.debug(f"{s.name} = {s.state}")

    for a in dir(atag.climate):
        _LOGGER.debug("atag.climate.%s = %r" % (a, getattr(atag.climate, a)))

    await atag.climate.set_preset_mode("manual")
    await atag.climate.set_temp(11)

    _LOGGER.debug(atag.report.report_time)
    _LOGGER.debug(atag.dhw.temperature)


asyncio.run(main())



"""Constants for ATAG API."""
CLASSES = {
    "temp": ["temperature", None, "mdi:thermometer"],
    "pres": ["pressure", "bar", "mdi:gauge"],
    "time": ["time", None, "mdi:clock"],
    "duration": ["duration", None, "mdi:timer"],
    "hours": [None, "h", "mdi:clock"],
}

SENSORS = {
    "burning_hours": "Burning Hours",
    #    "download_url": "API Version",
    "outside_temp": "Outside Temperature",
    "rel_mod_level": "Flame",
    "tout_avg": "Average Outside Temperature",
    "weather_status": "Weather Status",
    #    "ch_control_mode": "CH HVAC Mode",
    #    "ch_mode": "CH Preset Mode",
    #    "ch_mode_duration": "CH Preset Duration",
    #    "ch_mode_temp": "CH Target Temperature",
    "ch_return_temp": "CH Return Temperature",
    "ch_water_pres": "CH Water Pressure",
    "ch_water_temp": "CH Water Temperature",
}

STATES = {
    "weather_status": {
        0: {"state": "Sunny", "icon": "mdi:weather-sunny"},
        1: {"state": "Clear", "icon": "mdi:weather-night"},
        2: {"state": "Rainy", "icon": "mdi:weather-rainy"},
        3: {"state": "Snowy", "icon": "mdi:weather-snowy"},
        4: {"state": "Haily", "icon": "mdi:weather-hail"},
        5: {"state": "Windy", "icon": "mdi:weather-windy"},
        6: {"state": "Misty", "icon": "mdi:weather-fog"},
        7: {"state": "Cloudy", "icon": "mdi:weather-cloudy"},
        8: {"state": "Partly Sunny", "icon": "mdi:weather-partly-cloudy"},
        9: {"state": "Partly Cloudy", "icon": "mdi:cloud"},
        10: {"state": "Shower", "icon": "mdi:weather-pouring"},
        11: {"state": "Lightning", "icon": "mdi:weather-lightning"},
        12: {"state": "Hurricane", "icon": "mdi:weather-hurricane"},
        13: {"state": "Unknown", "icon": "mdi:cloud-question"},
    },
    "temp_unit": {0: "°C", 1: "°F"},
    "ch_mode": {1: "Manual", 2: "Auto", 3: "away", 4: "Extend", 5: "boost"},
    "ch_control_mode": {0: "heat", 1: "auto"},
}
DEFAULT_PORT = 10000



"""Automatic discovery of ATAG Thermostat on LAN."""
import asyncio
from .helpers import RequestError

ATAG_UDP_PORT = 11000
LOCALHOST = "0.0.0.0"


class Discovery(asyncio.DatagramProtocol):
    """Discovery class."""

    def __init__(self):
        self.data = asyncio.Future()

    def connection_made(self, transport):
        print("Listening on UDP {}".format(ATAG_UDP_PORT))

    def datagram_received(self, data, addr):
        self.data.set_result([data, addr])


async def async_discover_atag():
    """Discover Atag on local network."""
    # return format: [b'ONE xxxx-xxxx-xxxx_xx-xx-xxx-xxx (ST)',
    # ('xxx.xxx.x.x', xxxx)]
    trans, proto = await asyncio.get_event_loop().create_datagram_endpoint(
        Discovery, local_addr=(LOCALHOST, ATAG_UDP_PORT)
    )
    try:
        result = await asyncio.wait_for(proto.data, timeout=30)
        host_ip = result[1][0]
        device_id = result[0].decode().split()[1]
        trans.close()
    except asyncio.TimeoutError:
        trans.close()
        raise RequestError("Host discovery failed")
    return host_ip, device_id


def discover_atag():
    """Discover Atag on local network."""
    # return format: [b'ONE xxxx-xxxx-xxxx_xx-xx-xxx-xxx (ST)',
    # ('xxx.xxx.x.x', xxxx)]
    import socket

    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.bind(("", 11000))
    try:
        while True:
            result = sock.recvfrom(37)
            host_ip = result[1][0]
            device_id = result[0].decode().split()[1]
    except socket.timeout:
        pass
    except Exception as err:
        raise RequestError(err)
    return host_ip, device_id



"""Prepare for Pypi."""

from distutils.core import setup

setup(
    name="pyatag",  # How you named your package folder (MyLib)
    packages=["pyatag"],  # Chose the same as "name"
    version="0.3.1.2",  # Sta2t with a small number
    license="MIT",
    description="Connection to ATAG One Thermostat",
    author="Mats",  # Type in your name
    author_email="mats.nelissen@gmail.com",  # Type in your E-Mail
    url="https://github.com/MatsNl/pyatag",
    install_requires=["asyncio", "aiohttp",],
    classifiers=[
        "Development Status :: 3 - Alpha",
        #    'Intended Audience :: Developers',    # Define developer audience
        #    'Topic :: Software Development :: Build Tools',
        "License :: OSI Approved :: MIT License",  # Again, pick a license
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
)
